import io
import logging
import os
from itertools import repeat, starmap
from threading import Thread
from typing import Callable, Optional

from .backends.base import STREAM_TRANSFER_IN_PROGRESS, STREAM_WAITING_TO_START

CHUNK_SIZE = 4096

log = logging.getLogger(__name__)


def repeatfunc(
    func: Callable[..., None], times: Optional[int] = None, *args
):  # from the itertools receipes
    """Repeat calls to func with specified arguments.

    Example:  repeatfunc(random.random)

    :param args: params to the function to call.
    :param times: number of times to repeat.
    :param func:  the function to repeatedly call.
    """
    if times is None:
        return starmap(func, repeat(args))
    return starmap(func, repeat(args, times))


class Tee:
    """Tee implements a multi reader / single writer"""

    def __init__(self, incoming_stream, clients):
        """clients is a list of objects implementing callback_stream"""
        self.incoming_stream = incoming_stream
        self.clients = clients

    def start(self) -> Thread:
        """starts the transfer asynchronously"""
        t = Thread(target=self.run)
        t.start()
        return t

    def run(self):
        """streams to all the clients synchronously"""
        nb_clients = len(self.clients)
        pipes = [
            (io.open(r, "rb"), io.open(w, "wb"))
            for r, w in repeatfunc(os.pipe, nb_clients)
        ]
        streams = [self.incoming_stream.clone(pipe[0]) for pipe in pipes]

        def streamer(index):
            try:
                self.clients[index].callback_stream(streams[index])
                if streams[index].status == STREAM_WAITING_TO_START:
                    streams[index].reject()
                    plugin = self.clients[index].name
                    logging.warning(
                        "%s did not accept nor reject the incoming file transfer",
                        plugin,
                    )
                    logging.warning("I reject it as a fallback.")
            except Exception as _:
                # internal error, mark the error.
                streams[index].error()
            else:
                if streams[index].status == STREAM_TRANSFER_IN_PROGRESS:
                    # if the plugin didn't do it by itself, mark the transfer as a success.
                    streams[index].success()
            # stop the stream if the callback_stream returns
            read, write = pipes[index]
            pipes[index] = (None, None)  # signal the main thread to stop streaming
            read.close()
            write.close()

        threads = [Thread(target=streamer, args=(i,)) for i in range(nb_clients)]

        for thread in threads:
            thread.start()

        while True:
            if self.incoming_stream.closed:
                break
            chunk = self.incoming_stream.read(CHUNK_SIZE)
            log.debug("dispatch %d bytes", len(chunk))
            if not chunk:
                break
            for _, w in pipes:
                if w:
                    w.write(chunk)
        log.debug("EOF detected")
        for r, w in pipes:
            if w:
                w.close()  # close should flush too
        # we want to be sure that if we join on the main thread,
        # everything is either fully transfered or errored
        for thread in threads:
            thread.join()



from errbot import BotPlugin, botcmd

INROOM, USER, MULTILINE = "inroom", "user", "multiline"


class TextModeCmds(BotPlugin):
    """
    Internal to TextBackend.
    """

    __errdoc__ = "Added commands for testing purposes"

    def activate(self):
        # This won't activate the plugin in anything else than text mode.
        if self.mode != "text":
            return

        super().activate()

        # Some defaults if it was never used before'.
        if INROOM not in self:
            self[INROOM] = False

        if USER not in self:
            self[USER] = self.build_identifier(self.bot_config.BOT_ADMINS[0])

        if MULTILINE not in self:
            self[MULTILINE] = False

        # Restore the values to their live state.
        self._bot._inroom = self[INROOM]
        self._bot.user = self[USER]
        self._bot._multiline = self[MULTILINE]

    def deactivate(self):
        # Save the live state.
        self[INROOM] = self._bot._inroom
        self[USER] = self._bot.user
        self[MULTILINE] = self._bot._multiline

        super().deactivate()

    @botcmd
    def inroom(self, msg, args):
        """
        This puts you in a room with the bot.
        """
        self._bot._inroom = True
        if args:
            room = args
        else:
            room = "#testroom"
        self._bot.query_room(room).join()
        return f"Joined Room {room}."

    @botcmd
    def inperson(self, msg, _):
        """
        This puts you in a 1-1 chat with the bot.
        """
        self._bot._inroom = False
        return "Now in one-on-one with the bot."

    @botcmd
    def asuser(self, msg, args):
        """
        This puts you in a room with the bot. You can specify a name otherwise it will default to 'luser'.
        """
        if args:
            usr = args
            if usr[0] != "@":
                usr = "@" + usr
            self._bot.user = self.build_identifier(usr)
        else:
            self._bot.user = self.build_identifier("@luser")
        return f"You are now: {self._bot.user}."

    @botcmd
    def asadmin(self, msg, _):
        """
        This puts you in a 1-1 chat with the bot.
        """
        self._bot.user = self.build_identifier(self.bot_config.BOT_ADMINS[0])
        return f"You are now an admin: {self._bot.user}."

    @botcmd
    def ml(self, msg, _):
        """
        Switch back and forth between normal mode and multiline mode. Use this if you want to test
        commands spanning multiple lines. Note: in multiline, press enter twice to end and send the message.
        """
        self._bot._multiline = not self._bot._multiline
        return (
            "Multiline mode, press enter twice to end messages"
            if self._bot._multiline
            else "Normal one line mode."
        )



import fnmatch

from errbot import BotPlugin, cmdfilter
from errbot.backends.base import RoomOccupant

BLOCK_COMMAND = (None, None, None)


def get_acl_usr(msg):
    """Return the ACL attribute of the sender of the given message"""
    if hasattr(
        msg.frm, "aclattr"
    ):  # if the identity requires a special field to be used for acl
        return msg.frm.aclattr
    return msg.frm.person  # default


def get_acl_room(room):
    """Return the ACL attribute of the room used for a given message"""
    if hasattr(room, "aclattr"):
        return room.aclattr
    return str(room)  # old behaviour


def glob(text, patterns):
    """
    Match text against the list of patterns according to unix glob rules.
    Return True if a match is found, False otherwise.
    """
    if isinstance(patterns, str):
        patterns = (patterns,)
    if not isinstance(text, str):
        text = str(text)
    return any(fnmatch.fnmatchcase(text, str(pattern)) for pattern in patterns)


def ciglob(text, patterns):
    """
    Case-insensitive version of glob.

    Match text against the list of patterns according to unix glob rules.
    Return True if a match is found, False otherwise.
    """
    if isinstance(patterns, str):
        patterns = (patterns,)
    return glob(text.lower(), [p.lower() for p in patterns])


class ACLS(BotPlugin):
    """
    This plugin implements access controls for commands, allowing them to be
    restricted via various rules.
    """

    def access_denied(self, msg, reason, dry_run):
        if not dry_run and not self.bot_config.HIDE_RESTRICTED_ACCESS:
            self._bot.send_simple_reply(msg, reason)
        return BLOCK_COMMAND

    @cmdfilter
    def acls(self, msg, cmd, args, dry_run):
        """
        Check command against ACL rules as defined in the bot configuration.

        :param msg: The original chat message.
        :param cmd: The command name itself.
        :param args: Arguments passed to the command.
        :param dry_run: True when this is a dry-run.
        """
        self.log.debug("Check %s for ACLs.", cmd)
        f = self._bot.all_commands[cmd]
        cmd_str = f"{f.__self__.name}:{cmd}"

        usr = get_acl_usr(msg)
        acl = self.bot_config.ACCESS_CONTROLS_DEFAULT.copy()
        for pattern, acls in self.bot_config.ACCESS_CONTROLS.items():
            if ":" not in pattern:
                pattern = f"*:{pattern}"
            if ciglob(cmd_str, (pattern,)):
                acl.update(acls)
                break

        self.log.info(
            f"Matching ACL {acl} against username {usr} for command {cmd_str}."
        )
        if "allowargs" in acl and not glob(args, acl["allowargs"]):
            return self.access_denied(
                msg,
                "You're not allowed to access this command using the provided arguments",
                dry_run,
            )
        if "denyargs" in acl and glob(args, acl["denyargs"]):
            return self.access_denied(
                msg,
                "You're not allowed to access this command using the provided arguments",
                dry_run,
            )

        if "allowusers" in acl and not glob(usr, acl["allowusers"]):
            return self.access_denied(
                msg, "You're not allowed to access this command from this user", dry_run
            )
        if "denyusers" in acl and glob(usr, acl["denyusers"]):
            return self.access_denied(
                msg, "You're not allowed to access this command from this user", dry_run
            )
        if msg.is_group:
            if not isinstance(msg.frm, RoomOccupant):
                raise Exception(
                    f"msg.frm is not a RoomOccupant. Class of frm: {msg.frm.__class__}"
                )
            room = get_acl_room(msg.frm.room)
            if "allowmuc" in acl and acl["allowmuc"] is False:
                return self.access_denied(
                    msg,
                    "You're not allowed to access this command from a chatroom",
                    dry_run,
                )

            if "allowrooms" in acl and not glob(room, acl["allowrooms"]):
                return self.access_denied(
                    msg,
                    "You're not allowed to access this command from this room",
                    dry_run,
                )

            if "denyrooms" in acl and glob(room, acl["denyrooms"]):
                return self.access_denied(
                    msg,
                    "You're not allowed to access this command from this room",
                    dry_run,
                )
        elif "allowprivate" in acl and acl["allowprivate"] is False:
            return self.access_denied(
                msg,
                "You're not allowed to access this command via private message to me",
                dry_run,
            )

        self.log.debug(f"Check if {cmd} is admin only command.")
        if f._err_command_admin_only:
            if not glob(get_acl_usr(msg), self.bot_config.BOT_ADMINS):
                return self.access_denied(
                    msg, "This command requires bot-admin privileges", dry_run
                )
            # For security reasons, admin-only commands are direct-message only UNLESS
            # specifically overridden by setting allowmuc to True for such commands.
            if msg.is_group and not acl.get("allowmuc", False):
                return self.access_denied(
                    msg,
                    "This command may only be issued through a direct message",
                    dry_run,
                )

        return msg, cmd, args



#!/usr/bin/env python

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import argparse
import ast
import logging
import os
import sys
from os import W_OK, access, getcwd, path, sep
from pathlib import Path
from platform import system
from typing import Optional, Union

from errbot.bootstrap import CORE_BACKENDS
from errbot.logs import root_logger
from errbot.plugin_wizard import new_plugin_wizard
from errbot.utils import collect_roots, entry_point_plugins
from errbot.version import VERSION

log = logging.getLogger(__name__)


# noinspection PyUnusedLocal
def debug(sig, frame) -> None:
    """Interrupt running process, and provide a python prompt for
    interactive debugging."""
    d = {"_frame": frame}  # Allow access to frame object.
    d.update(frame.f_globals)  # Unless shadowed by global
    d.update(frame.f_locals)

    i = code.InteractiveConsole(d)
    message = "Signal received : entering python shell.\nTraceback:\n"
    message += "".join(traceback.format_stack(frame))
    i.interact(message)


ON_WINDOWS = system() == "Windows"

if not ON_WINDOWS:
    import code
    import signal
    import traceback

    from daemonize import Daemonize

    signal.signal(signal.SIGUSR1, debug)  # Register handler for debugging


def get_config(config_path: str):
    config_fullpath = config_path
    if not path.exists(config_fullpath):
        log.error(f"I cannot find the config file {config_path}.")
        log.error("You can change this path with the -c parameter see --help")
        log.info(
            f'You can use the template {os.path.realpath(os.path.join(__file__, os.pardir, "config-template.py"))}'
            f" as a base and copy it to {config_path}."
        )
        log.info("You can then customize it.")
        exit(-1)

    try:
        config = __import__(path.splitext(path.basename(config_fullpath))[0])
        log.info("Config check passed...")
        return config
    except Exception:
        log.exception(
            f"I could not import your config from {config_fullpath}, please check the error below..."
        )
        exit(-1)


def _read_dict() -> dict:
    from collections.abc import Mapping

    new_dict = ast.literal_eval(sys.stdin.read())
    if not isinstance(new_dict, Mapping):
        raise ValueError(
            f"A dictionary written in python is needed from stdin. "
            f"Type={type(new_dict)}, Value = {repr(new_dict)}."
        )
    return new_dict


def main() -> None:
    execution_dir = getcwd()

    # By default insert the execution path (useful to be able to execute Errbot from
    # the source tree directly without installing it.
    sys.path.insert(0, execution_dir)

    parser = argparse.ArgumentParser(description="The main entry point of the errbot.")
    parser.add_argument(
        "-c",
        "--config",
        default=None,
        help="Full path to your config.py (default: config.py in current working directory).",
    )

    mode_selection = parser.add_mutually_exclusive_group()
    mode_selection.add_argument(
        "-v", "--version", action="version", version=f"Errbot version {VERSION}"
    )
    mode_selection.add_argument(
        "-r",
        "--restore",
        nargs="?",
        default=None,
        const="default",
        help="restore a bot from backup.py (default: backup.py from the bot data directory)",
    )
    mode_selection.add_argument(
        "-l", "--list", action="store_true", help="list all available backends"
    )
    mode_selection.add_argument(
        "--new-plugin",
        nargs="?",
        default=None,
        const="current_dir",
        help="create a new plugin in the specified directory",
    )
    mode_selection.add_argument(
        "-i",
        "--init",
        nargs="?",
        default=None,
        const=".",
        help="Initialize a simple bot minimal configuration in the optionally "
        "given directory (otherwise it will be the working directory). "
        "This will create a data subdirectory for the bot data dir and a plugins directory"
        " for your plugin development with an example in it to get you started.",
    )
    # storage manipulation
    mode_selection.add_argument(
        "--storage-set",
        nargs=1,
        help="DANGER: Delete the given storage namespace "
        "and set the python dictionary expression "
        "passed on stdin.",
    )
    mode_selection.add_argument(
        "--storage-merge",
        nargs=1,
        help="DANGER: Merge in the python dictionary expression "
        "passed on stdin into the given storage namespace.",
    )
    mode_selection.add_argument(
        "--storage-get",
        nargs=1,
        help="Dump the given storage namespace in a "
        "format compatible for --storage-set and "
        "--storage-merge.",
    )

    mode_selection.add_argument(
        "-T",
        "--text",
        dest="backend",
        action="store_const",
        const="Text",
        help="force local text backend",
    )

    if not ON_WINDOWS:
        option_group = parser.add_argument_group("optional daemonization arguments")
        option_group.add_argument(
            "-d",
            "--daemon",
            action="store_true",
            help="Detach the process from the console",
        )
        option_group.add_argument(
            "-p",
            "--pidfile",
            default=None,
            help="Specify the pid file for the daemon (default: current bot data directory)",
        )

    args = vars(parser.parse_args())  # create a dictionary of args

    if args["init"]:
        try:
            import pathlib
            import shutil

            import jinja2

            base_dir = (
                pathlib.Path.cwd()
                if args["init"] == "."
                else Path(args["init"]).resolve()
            )

            if not base_dir.exists():
                print(f"Target directory {base_dir} must exist. Please create it.")

            data_dir = base_dir / "data"
            extra_plugin_dir = base_dir / "plugins"
            extra_backend_plugin_dir = base_dir / "backend-plugins"
            example_plugin_dir = extra_plugin_dir / "err-example"
            log_path = base_dir / "errbot.log"

            templates_dir = Path(os.path.dirname(__file__)) / "templates" / "initdir"
            env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(str(templates_dir)), autoescape=True
            )
            config_template = env.get_template("config.py.tmpl")

            data_dir.mkdir(exist_ok=True)
            extra_plugin_dir.mkdir(exist_ok=True)
            extra_backend_plugin_dir.mkdir(exist_ok=True)
            example_plugin_dir.mkdir(exist_ok=True)

            with open(base_dir / "config.py", "w") as f:
                f.write(
                    config_template.render(
                        data_dir=str(data_dir),
                        extra_plugin_dir=str(extra_plugin_dir),
                        extra_backend_plugin_dir=str(extra_backend_plugin_dir),
                        log_path=str(log_path),
                    )
                )

            shutil.copyfile(
                templates_dir / "example.plug", example_plugin_dir / "example.plug"
            )
            shutil.copyfile(
                templates_dir / "example.py", example_plugin_dir / "example.py"
            )
            print("Your Errbot directory has been correctly initialized!")
            if base_dir == pathlib.Path.cwd():
                print('Just do "errbot" and it should start in text/development mode.')
            else:
                print(
                    f'Just do "cd {args["init"]}" then "errbot" and it should start in text/development mode.'
                )
            sys.exit(0)
        except Exception as e:
            print(f"The initialization of your errbot directory failed: {e}.")
            sys.exit(1)

    # This must come BEFORE the config is loaded below, to avoid printing
    # logs as a side effect of config loading.
    if args["new_plugin"]:
        directory = (
            os.getcwd() if args["new_plugin"] == "current_dir" else args["new_plugin"]
        )
        for handler in logging.getLogger().handlers:
            root_logger.removeHandler(handler)
        try:
            new_plugin_wizard(directory)
        except KeyboardInterrupt:
            sys.exit(1)
        except Exception as e:
            sys.stderr.write(str(e) + "\n")
            sys.exit(1)
        finally:
            sys.exit(0)

    config_path = args["config"]
    # setup the environment to be able to import the config.py
    if config_path:
        # appends the current config in order to find config.py
        sys.path.insert(0, path.dirname(path.abspath(config_path)))
    else:
        config_path = execution_dir + sep + "config.py"

    config = get_config(config_path)  # will exit if load fails

    # Extra backend is expected to be a list type, convert string to list.
    extra_backend = getattr(config, "BOT_EXTRA_BACKEND_DIR", [])
    if isinstance(extra_backend, str):
        extra_backend = [extra_backend]
    ep = entry_point_plugins(group="errbot.backend_plugins")
    extra_backend.extend(ep)

    if args["list"]:
        from errbot.backend_plugin_manager import enumerate_backend_plugins

        print("Available backends:")
        roots = [CORE_BACKENDS] + extra_backend
        for backend in enumerate_backend_plugins(collect_roots(roots)):
            print(f"\t\t{backend.name}")
        sys.exit(0)

    def storage_action(namespace, fn):
        # Used to defer imports until it is really necessary during the loading time.
        from errbot.bootstrap import get_storage_plugin
        from errbot.storage import StoreMixin

        try:
            with StoreMixin() as sdm:
                sdm.open_storage(get_storage_plugin(config), namespace)
                fn(sdm)
            return 0
        except Exception as e:
            print(str(e), file=sys.stderr)
            return -3

    if args["storage_get"]:

        def p(sdm):
            print(repr(dict(sdm)))

        err_value = storage_action(args["storage_get"][0], p)
        sys.exit(err_value)

    if args["storage_set"]:

        def replace(sdm):
            new_dict = (
                _read_dict()
            )  # fail early and don't erase the storage if the input is invalid.
            sdm.clear()
            sdm.update(new_dict)

        err_value = storage_action(args["storage_set"][0], replace)
        sys.exit(err_value)

    if args["storage_merge"]:

        def merge(sdm):
            from deepmerge import always_merger

            new_dict = _read_dict()
            for key, value in new_dict.items():
                with sdm.mutable(key, {}) as conf:
                    always_merger.merge(conf, value)

        err_value = storage_action(args["storage_merge"][0], merge)
        sys.exit(err_value)

    if args["restore"]:
        backend = "Null"  # we don't want any backend when we restore
    elif args["backend"] is None:
        if not hasattr(config, "BACKEND"):
            log.fatal("The BACKEND configuration option is missing in config.py")
            sys.exit(1)
        backend = config.BACKEND
    else:
        backend = args["backend"]

    log.info(f"Selected backend {backend}.")

    # Check if at least we can start to log something before trying to start
    # the bot (esp. daemonize it).

    log.info(f"Checking for {config.BOT_DATA_DIR}...")
    if not path.exists(config.BOT_DATA_DIR):
        raise Exception(
            f'The data directory "{config.BOT_DATA_DIR}" for the bot does not exist.'
        )
    if not access(config.BOT_DATA_DIR, W_OK):
        raise Exception(
            f'The data directory "{config.BOT_DATA_DIR}" should be writable for the bot.'
        )

    if (not ON_WINDOWS) and args["daemon"]:
        if args["backend"] == "Text":
            raise Exception("You cannot run in text and daemon mode at the same time")
        if args["restore"]:
            raise Exception("You cannot restore a backup in daemon mode.")
        if args["pidfile"]:
            pid = args["pidfile"]
        else:
            pid = config.BOT_DATA_DIR + sep + "err.pid"

        # noinspection PyBroadException
        try:

            def action():
                from errbot.bootstrap import bootstrap

                bootstrap(backend, root_logger, config)

            daemon = Daemonize(app="err", pid=pid, action=action, chdir=os.getcwd())
            log.info("Daemonizing")
            daemon.start()
        except Exception:
            log.exception("Failed to daemonize the process")
        exit(0)
    from errbot.bootstrap import bootstrap

    restore = args["restore"]
    if restore == "default":  # restore with no argument, get the default location
        restore = path.join(config.BOT_DATA_DIR, "backup.py")

    bootstrap(backend, root_logger, config, restore)
    log.info("Process exiting")


if __name__ == "__main__":
    main()



import inspect
import logging
import sys
from typing import Optional

COLORS = {
    "DEBUG": "cyan",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "red",
}

NO_COLORS = {
    "DEBUG": "",
    "INFO": "",
    "WARNING": "",
    "ERROR": "",
    "CRITICAL": "",
}


def ispydevd():
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            return True
    return False


root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)

pydev = ispydevd()
stream = sys.stdout if pydev else sys.stderr
isatty = (
    pydev or stream.isatty()
)  # force isatty if we are under pydev because it supports coloring anyway.
console_hdlr = logging.StreamHandler(stream)


def get_log_colors(theme_color: Optional[str] = None) -> str:
    """Return a tuple containing the log format string and a log color dict"""
    if theme_color == "light":
        text_color_theme = "white"
    elif theme_color == "dark":
        text_color_theme = "black"
    else:  # Anything else produces nocolor
        return "%(name)-25.25s%(reset)s %(message)s%(reset)s", NO_COLORS

    return f"%(name)-25.25s%(reset)s %({text_color_theme})s%(message)s%(reset)s", COLORS


def format_logs(
    formatter: Optional[logging.Formatter] = None, theme_color: Optional[str] = None
) -> None:
    """
    You may either use the formatter parameter to provide your own
    custom formatter, or the theme_color parameter to use the
    built in color scheme formatter.
    """
    if formatter:
        console_hdlr.setFormatter(formatter)
    # if isatty and not True:
    elif isatty:
        from colorlog import ColoredFormatter  # noqa

        log_format, colors_dict = get_log_colors(theme_color)
        color_formatter = ColoredFormatter(
            "%(asctime)s %(log_color)s%(levelname)-8s%(reset)s " + log_format,
            datefmt="%H:%M:%S",
            reset=True,
            log_colors=colors_dict,
        )
        console_hdlr.setFormatter(color_formatter)
    else:
        console_hdlr.setFormatter(
            logging.Formatter("%(asctime)s %(levelname)-8s %(name)-25s %(message)s")
        )
    root_logger.addHandler(console_hdlr)



import logging
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from errbot.plugin_info import PluginInfo

log = logging.getLogger(__name__)


def make_templates_path(root: Path) -> Path:
    return root / "templates"


system_templates_path = str(make_templates_path(Path(__file__).parent))
template_path = [system_templates_path]
env = Environment(
    loader=FileSystemLoader(template_path),
    trim_blocks=True,
    keep_trailing_newline=False,
    autoescape=True,
)


def tenv() -> Environment:
    return env


def add_plugin_templates_path(plugin_info: PluginInfo) -> None:
    global env
    tmpl_path = make_templates_path(plugin_info.location.parent)
    if tmpl_path.exists():
        log.debug(
            "Templates directory found for %s plugin [%s]", plugin_info.name, tmpl_path
        )
        template_path.append(str(tmpl_path))  # for webhooks

        # Ditch and recreate a new templating environment
        env = Environment(loader=FileSystemLoader(template_path), autoescape=True)
        return
    log.debug(
        "No templates directory found for %s plugin in [%s]",
        plugin_info.name,
        tmpl_path,
    )


def remove_plugin_templates_path(plugin_info: PluginInfo) -> None:
    global env
    tmpl_path = str(make_templates_path(plugin_info.location.parent))
    if tmpl_path in template_path:
        template_path.remove(tmpl_path)
        # Ditch and recreate a new templating environment
        env = Environment(loader=FileSystemLoader(template_path), autoescape=True)



import logging
import os
import shutil
from ast import literal_eval
from pprint import pformat

from errbot import BotPlugin, botcmd
from errbot.plugin_manager import (
    PluginActivationException,
    PluginConfigurationException,
)
from errbot.repo_manager import RepoException


class Plugins(BotPlugin):
    @botcmd(admin_only=True)
    def repos_install(self, _, args):
        """install a plugin repository from the given source or a known public repo (see !repos to find those).
        for example from a known repo: !install err-codebot
        for example a git url: git@github.com:gbin/plugin.git
        or an url towards an archive: https://github.com/errbotio/err-helloworld/archive/refs/heads/master.zip
        """
        args = args.strip()
        if not args:
            yield 'Please specify a repository listed in "!repos" or "give me the URL to a git repository that I should clone for you."'
            return
        try:
            yield f"Installing {args}..."
            local_path = self._bot.repo_manager.install_repo(args)
            errors = self._bot.plugin_manager.update_plugin_places(
                self._bot.repo_manager.get_all_repos_paths()
            )
            if errors:
                v = "\n".join(errors.values())
                yield f"Some plugins are generating errors:\n{v}."
                # if the load of the plugin failed, uninstall cleanly teh repo
                for path in errors.keys():
                    if str(path).startswith(local_path):
                        yield f"Removing {local_path} as it did not load correctly."
                        shutil.rmtree(local_path)
            else:
                yield f"A new plugin repository has been installed correctly from {args}. Refreshing the plugins commands..."
            loading_errors = self._bot.plugin_manager.activate_non_started_plugins()
            if loading_errors:
                yield loading_errors
            yield "Plugins reloaded."
        except RepoException as re:
            yield f"Error installing the repo: {re}"

    @botcmd(admin_only=True)
    def repos_uninstall(self, _, repo_name):
        """uninstall a plugin repository by name."""
        if not repo_name.strip():
            yield "You should have a repo name as argument"
            return

        repos = self._bot.repo_manager.get_installed_plugin_repos()

        if repo_name not in repos:
            yield f"This repo is not installed check with {self._bot.prefix}repos the list of installed ones"
            return

        plugin_path = os.path.join(self._bot.repo_manager.plugin_dir, repo_name)
        self._bot.plugin_manager.remove_plugins_from_path(plugin_path)
        self._bot.repo_manager.uninstall_repo(repo_name)
        yield f"Repo {repo_name} removed."

    @botcmd(template="repos")
    def repos(self, _, args):
        """list the current active plugin repositories"""

        installed_repos = self._bot.repo_manager.get_installed_plugin_repos()

        all_names = [name for name in installed_repos]

        repos = {"repos": []}

        for repo_name in all_names:
            installed = False

            if repo_name in installed_repos:
                installed = True

            from_index = self._bot.repo_manager.get_repo_from_index(repo_name)

            if from_index is not None:
                description = "\n".join(
                    (f"{plug.name}: {plug.documentation}" for plug in from_index)
                )
            else:
                description = "No description."

            # installed, public, name, desc
            repos["repos"].append(
                (installed, from_index is not None, repo_name, description)
            )

        return repos

    @botcmd(template="repos2")
    def repos_search(self, _, args):
        """Searches the repo index.
        for example: !repos search jenkins
        """
        if not args:
            # TODO(gbin): return all the repos.
            return {"error": "Please specify a keyword."}
        return {"repos": self._bot.repo_manager.search_repos(args)}

    @botcmd(split_args_with=" ", admin_only=True)
    def repos_update(self, _, args):
        """update the bot and/or plugins
        use: !repos update all
        to update everything
        or: !repos update repo_name repo_name ...
        to update selectively some repos
        """
        if "all" in args:
            results = self._bot.repo_manager.update_all_repos()
        else:
            results = self._bot.repo_manager.update_repos(args)

        yield "Start updating ... "

        for d, success, feedback in results:
            if success:
                yield f"Update of {d} succeeded...\n\n{feedback}\n\n"

                for plugin in self._bot.plugin_manager.get_plugins_by_path(d):
                    if hasattr(plugin, "is_activated") and plugin.is_activated:
                        name = plugin.name
                        yield f"/me is reloading plugin {name}"
                        try:
                            self._bot.plugin_manager.reload_plugin_by_name(plugin.name)
                            yield f"Plugin {plugin.name} reloaded."
                        except PluginActivationException as pae:
                            yield f"Error reactivating plugin {plugin.name}: {pae}"
            else:
                yield f"Update of {d} failed...\n\n{feedback}"

        yield "Done."

    @botcmd(split_args_with=" ", admin_only=True)
    def plugin_config(self, _, args):
        """configure or get the configuration / configuration template for a specific plugin
        ie.
        !plugin config ExampleBot
        could return a template if it is not configured:
        {'LOGIN': 'example@example.com', 'PASSWORD': 'password', 'DIRECTORY': '/toto'}
        Copy paste, adapt so can configure the plugin:
        !plugin config ExampleBot {'LOGIN': 'my@email.com', 'PASSWORD': 'myrealpassword', 'DIRECTORY': '/tmp'}
        It will then reload the plugin with this config.
        You can at any moment retrieve the current values:
        !plugin config ExampleBot
        should return:
        {'LOGIN': 'my@email.com', 'PASSWORD': 'myrealpassword', 'DIRECTORY': '/tmp'}
        """
        plugin_name = args[0]
        if self._bot.plugin_manager.is_plugin_blacklisted(plugin_name):
            return f"Load this plugin first with {self._bot.prefix} load {plugin_name}."
        obj = self._bot.plugin_manager.get_plugin_obj_by_name(plugin_name)
        if obj is None:
            return f"Unknown plugin or the plugin could not load {plugin_name}."
        template_obj = obj.get_configuration_template()
        if template_obj is None:
            return "This plugin is not configurable."

        if len(args) == 1:
            response = (
                f"Default configuration for this plugin (you can copy and paste this directly as a command):"
                f"\n\n```\n{self._bot.prefix}plugin config {plugin_name} {pformat(template_obj)}\n```"
            )

            current_config = self._bot.plugin_manager.get_plugin_configuration(
                plugin_name
            )
            if current_config:
                response += (
                    f"\n\nCurrent configuration:\n\n```\n{self._bot.prefix}plugin config {plugin_name} "
                    f"{pformat(current_config)}\n```"
                )
            return response

        # noinspection PyBroadException
        try:
            real_config_obj = literal_eval(" ".join(args[1:]))
        except Exception:
            self.log.exception("Invalid expression for the configuration of the plugin")
            return "Syntax error in the given configuration"
        if type(real_config_obj) != type(template_obj):
            return "It looks fishy, your config type is not the same as the template!"

        self._bot.plugin_manager.set_plugin_configuration(plugin_name, real_config_obj)

        try:
            self._bot.plugin_manager.deactivate_plugin(plugin_name)
        except PluginActivationException as pae:
            return f"Error deactivating {plugin_name}: {pae}."

        try:
            self._bot.plugin_manager.activate_plugin(plugin_name)
        except PluginConfigurationException as ce:
            self.log.debug(
                "Invalid configuration for the plugin, reverting the plugin to unconfigured."
            )
            self._bot.plugin_manager.set_plugin_configuration(plugin_name, None)
            return f"Incorrect plugin configuration: {ce}."
        except PluginActivationException as pae:
            return f"Error activating plugin: {pae}."

        return "Plugin configuration done."

    def formatted_plugin_list(self, active_only=True):
        """
        Return a formatted, plain-text list of loaded plugins.

        When active_only=True, this will only return plugins which
        are actually active. Otherwise, it will also include inactive
        (blacklisted) plugins.
        """
        if active_only:
            all_plugins = self._bot.plugin_manager.get_all_active_plugin_names()
        else:
            all_plugins = self._bot.plugin_manager.get_all_plugin_names()
        return "\n".join(("- " + plugin for plugin in all_plugins))

    @botcmd(admin_only=True)
    def plugin_reload(self, _, args):
        """reload a plugin: reload the code of the plugin leaving the activation status intact."""
        name = args.strip()
        if not name:
            yield (
                f"Please tell me which of the following plugins to reload:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )
            return
        if name not in self._bot.plugin_manager.get_all_plugin_names():
            yield (
                f"{name} isn't a valid plugin name. "
                f"The current plugins are:\n{self.formatted_plugin_list(active_only=False)}"
            )
            return

        if name not in self._bot.plugin_manager.get_all_active_plugin_names():
            answer = f"Warning: plugin {name} is currently not activated. "
            answer += f"Use `{self._bot.prefix}plugin activate {name}` to activate it."
            yield answer

        try:
            self._bot.plugin_manager.reload_plugin_by_name(name)
            yield f"Plugin {name} reloaded."
        except PluginActivationException as pae:
            yield f"Error activating plugin {name}: {pae}."

    @botcmd(admin_only=True)
    def plugin_activate(self, _, args):
        """activate a plugin. [calls .activate() on the plugin]"""
        args = args.strip()
        if not args:
            return (
                f"Please tell me which of the following plugins to activate:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (
                f"{args} isn't a valid plugin name. The current plugins are:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )
        if args in self._bot.plugin_manager.get_all_active_plugin_names():
            return f"{args} is already activated."

        try:
            self._bot.plugin_manager.activate_plugin(args)
        except PluginActivationException as pae:
            return f"Error activating plugin: {pae}"
        return f"Plugin {args} activated."

    @botcmd(admin_only=True)
    def plugin_deactivate(self, _, args):
        """deactivate a plugin. [calls .deactivate on the plugin]"""
        args = args.strip()
        if not args:
            return (
                f"Please tell me which of the following plugins to deactivate:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (
                f"{args} isn't a valid plugin name. The current plugins are:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )
        if args not in self._bot.plugin_manager.get_all_active_plugin_names():
            return f"{args} is already deactivated."

        try:
            self._bot.plugin_manager.deactivate_plugin(args)
        except PluginActivationException as pae:
            return f"Error deactivating {args}: {pae}"
        return f"Plugin {args} deactivated."

    @botcmd(admin_only=True)
    def plugin_blacklist(self, _, args):
        """Blacklist a plugin so that it will not be loaded automatically during bot startup.
        If the plugin is currently activated, it will deactiveate it first."""
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (
                f"{args} isn't a valid plugin name. The current plugins are:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )

        if args in self._bot.plugin_manager.get_all_active_plugin_names():
            try:
                self._bot.plugin_manager.deactivate_plugin(args)
            except PluginActivationException as pae:
                return f"Error deactivating {args}: {pae}."
        return self._bot.plugin_manager.blacklist_plugin(args)

    @botcmd(admin_only=True)
    def plugin_unblacklist(self, _, args):
        """Remove a plugin from the blacklist"""
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (
                f"{args} isn't a valid plugin name. The current plugins are:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )

        if args not in self._bot.plugin_manager.get_all_active_plugin_names():
            try:
                self._bot.plugin_manager.activate_plugin(args)
            except PluginActivationException as pae:
                return f"Error activating plugin: {pae}"

        return self._bot.plugin_manager.unblacklist_plugin(args)

    @botcmd(admin_only=True, template="plugin_info")
    def plugin_info(self, _, args):
        """Gives you a more technical information about a specific plugin."""
        pm = self._bot.plugin_manager
        if args not in pm.get_all_plugin_names():
            return (
                f"{args} isn't a valid plugin name. The current plugins are:\n"
                f"{self.formatted_plugin_list(active_only=False)}"
            )
        return {
            "plugin_info": pm.plugin_infos[args],
            "plugin": pm.plugins[args],
            "logging": logging,
        }



from __future__ import annotations

import logging
import sys
from datetime import datetime
from functools import lru_cache
from time import sleep
from typing import Callable, List, Optional, Tuple

from errbot.backends.base import (
    AWAY,
    DND,
    OFFLINE,
    ONLINE,
    Identifier,
    Message,
    Person,
    Presence,
    Room,
    RoomNotJoinedError,
    RoomOccupant,
)
from errbot.core import ErrBot
from errbot.rendering import text, xhtml, xhtmlim

log = logging.getLogger(__name__)

try:
    from slixmpp import JID, ClientXMPP
    from slixmpp.exceptions import IqError
    from slixmpp.xmlstream import cert, resolver

except ImportError:
    log.exception("Could not start the XMPP backend")
    log.fatal(
        """
    If you intend to use the XMPP backend please install the support for XMPP with:
    pip install errbot[XMPP]
    """
    )
    sys.exit(-1)

# LRU to cache the JID queries.
IDENTIFIERS_LRU = 1024


class XMPPIdentifier(Identifier):
    """
    This class is the parent and the basic contract of all the ways the backends
    are identifying a person on their system.
    """

    def __init__(self, node, domain, resource):
        if not node:
            raise Exception("An XMPPIdentifier needs to have a node.")
        if not domain:
            raise Exception("An XMPPIdentifier needs to have a domain.")
        self._node = node
        self._domain = domain
        self._resource = resource
        self._email = ""

    @property
    def node(self) -> str:
        return self._node

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def resource(self) -> str:
        return self._resource

    @property
    def person(self) -> str:
        return self._node + "@" + self._domain

    @property
    def nick(self) -> str:
        return self._node

    @property
    def fullname(self) -> None:
        return None  # Not supported by default on XMPP.

    @property
    def email(self):
        return self._email

    @property
    def client(self):
        return self._resource

    def __str__(self):
        answer = self._node + "@" + self._domain  # don't call .person: see below
        if self._resource:
            answer += "/" + self._resource
        return answer

    def __unicode__(self):
        return str(self.__str__())

    def __eq__(self, other):
        if not isinstance(other, XMPPIdentifier):
            log.debug("Weird, you are comparing an XMPPIdentifier to a %s", type(other))
            return False
        return (
            self._domain == other._domain
            and self._node == other._node
            and self._resource == other._resource
        )


class XMPPPerson(XMPPIdentifier, Person):
    aclattr = XMPPIdentifier.person

    def __eq__(self, other):
        if not isinstance(other, XMPPPerson):
            log.debug("Weird, you are comparing an XMPPPerson to a %s", type(other))
            return False
        return self._domain == other._domain and self._node == other._node


class XMPPRoom(XMPPIdentifier, Room):
    def __init__(self, room_jid, bot: ErrBot):
        self._bot = bot
        self.xep0045 = self._bot.conn.client.plugin["xep_0045"]
        node, domain, resource = split_identifier(room_jid)
        super().__init__(node, domain, resource)

    def join(
        self, username: Optional[str] = None, password: Optional[str] = None
    ) -> None:
        """
        Join the room.

        If the room does not exist yet, this will automatically call
        :meth:`create` on it first.
        """
        room = str(self)
        self.xep0045.join_muc(room, username, password=password)
        self._bot.conn.add_event_handler(
            f"muc::{room}::got_online", self._bot.user_joined_chat
        )
        self._bot.conn.add_event_handler(
            f"muc::{room}::got_offline", self._bot.user_left_chat
        )
        self.configure()
        self._bot.callback_room_joined(self, self._bot.bot_identifier)
        log.info("Joined room %s.", room)

    def leave(self, reason: Optional[str] = None) -> None:
        """
        Leave the room.

        :param reason:
            An optional string explaining the reason for leaving the room
        """
        if reason is None:
            reason = ""
        room = str(self)
        try:
            self.xep0045.leave_muc(
                room=room, nick=self.xep0045.ourNicks[room], msg=reason
            )

            self._bot.conn.del_event_handler(
                f"muc::{room}::got_online", self._bot.user_joined_chat
            )
            self._bot.conn.del_event_handler(
                f"muc::{room}::got_offline", self._bot.user_left_chat
            )
            log.info("Left room %s.", room)
            self._bot.callback_room_left(self, self._bot.bot_identifier)
        except KeyError:
            log.debug("Trying to leave %s while not in this room.", room)

    def create(self) -> None:
        """
        Not supported on this back-end (Slixmpp doesn't support it).
        Will join the room to ensure it exists, instead.
        """
        logging.warning(
            "XMPP back-end does not support explicit creation, joining room "
            "instead to ensure it exists."
        )
        self.join(username=str(self))

    def destroy(self) -> None:
        """
        Destroy the room.

        Calling this on a non-existing room is a no-op.
        """
        self.xep0045.destroy(str(self))
        log.info("Destroyed room %s.", self)

    @property
    def exists(self) -> bool:
        """
        Boolean indicating whether this room already exists or not.

        :getter:
            Returns `True` if the room exists, `False` otherwise.
        """
        logging.warning(
            "XMPP back-end does not support determining if a room exists. Returning the result of joined instead."
        )
        return self.joined

    @property
    def joined(self) -> bool:
        """
        Boolean indicating whether this room has already been joined.

        :getter:
            Returns `True` if the room has been joined, `False` otherwise.
        """
        return str(self) in self.xep0045.get_joined_rooms()

    @property
    def topic(self) -> Optional[str]:
        """
        The room topic.

        :getter:
            Returns the topic (a string) if one is set, `None` if no
            topic has been set at all.
        :raises:
            :class:`~RoomNotJoinedError` if the room has not yet been joined.
        """
        if not self.joined:
            raise RoomNotJoinedError("Must be in a room in order to see the topic.")
        try:
            return self._bot._room_topics[str(self)]
        except KeyError:
            return None

    @topic.setter
    def topic(self, topic: str) -> None:
        """
        Set the room's topic.

        :param topic:
            The topic to set.
        """
        # Not supported by Slixmpp at the moment :(
        raise NotImplementedError(
            "Setting the topic is not supported on this back-end."
        )

    @property
    def occupants(self) -> List[XMPPRoomOccupant]:
        """
        The room's occupants.

        :getter:
            Returns a list of :class:`~errbot.backends.base.MUCOccupant` instances.
        :raises:
            :class:`~MUCNotJoinedError` if the room has not yet been joined.
        """
        occupants = []
        try:
            for occupant in self.xep0045.rooms[str(self)].values():
                room_node, room_domain, _ = split_identifier(occupant["room"])
                nick = occupant["nick"]
                occupants.append(XMPPRoomOccupant(room_node, room_domain, nick, self))
        except KeyError:
            raise RoomNotJoinedError("Must be in a room in order to see occupants.")
        return occupants

    def invite(self, *args) -> None:
        """
        Invite one or more people into the room.

        :*args:
            One or more JID's to invite into the room.
        """
        room = str(self)
        for jid in args:
            self.xep0045.invite(room, jid)
            log.info("Invited %s to %s.", jid, room)

    def configure(self) -> None:
        """
        Configure the room.

        Currently this simply sets the default room configuration as
        received by the server. May be extended in the future to set
        a custom room configuration instead.
        """
        room = str(self)
        affiliation = None
        while affiliation is None:
            sleep(0.5)
            affiliation = self.xep0045.get_jid_property(
                room=room, nick=self.xep0045.our_nicks[room], jid_property="affiliation"
            )

        if affiliation == "owner":
            log.debug("Configuring room %s: we have owner affiliation.", room)
            form = yield from self.xep0045.get_room_config(room)
            self.xep0045.configure_room(room, form)
        else:
            log.debug(
                "Not configuring room %s: we don't have owner affiliation (affiliation=%s)",
                room,
                affiliation,
            )


class XMPPRoomOccupant(XMPPPerson, RoomOccupant):
    def __init__(self, node, domain, resource, room):
        super().__init__(node, domain, resource)
        self._room = room

    @property
    def person(self):
        return str(self)  # this is the full identifier.

    @property
    def real_jid(self) -> str:
        """
        The JID of the room occupant, they used to login.
        Will only work if the errbot is moderator in the MUC or it is not anonymous.
        """
        room_jid = self._node + "@" + self._domain
        jid = JID(self._room.xep0045.get_jid_property(room_jid, self.resource, "jid"))
        return jid.bare

    @property
    def room(self) -> XMPPRoom:
        return self._room

    nick = XMPPPerson.resource


class XMPPConnection:
    def __init__(
        self,
        jid,
        password,
        feature=None,
        keepalive=None,
        ca_cert=None,
        server=None,
        use_ipv6=None,
        bot=None,
        ssl_version=None,
    ):
        if feature is None:
            feature = {}
        self._bot = bot
        self.connected = False
        self.server = server
        self.client = ClientXMPP(
            jid, password, plugin_config={"feature_mechanisms": feature}
        )
        self.client.register_plugin("xep_0030")  # Service Discovery
        self.client.register_plugin("xep_0045")  # Multi-User Chat
        self.client.register_plugin("xep_0199")  # XMPP Ping
        self.client.register_plugin("xep_0203")  # XMPP Delayed messages
        self.client.register_plugin("xep_0249")  # XMPP direct MUC invites

        if keepalive is not None:
            self.client.whitespace_keepalive = (
                True  # Just in case Slixmpp's default changes to False in the future
            )
            self.client.whitespace_keepalive_interval = keepalive

        if use_ipv6 is not None:
            self.client.use_ipv6 = use_ipv6

        self.client.ca_certs = ca_cert  # Used for TLS certificate validation

        self.client.add_event_handler("session_start", self.session_start)

    def session_start(self, _):
        self.client.send_presence()
        self.client.get_roster()

    def connect(self) -> XMPPConnection:
        if not self.connected:
            if self.server is not None:
                self.client.connect(self.server)
            else:
                self.client.connect()
            self.connected = True
        return self

    def disconnect(self) -> None:
        self.client.disconnect(wait=True)
        self.connected = False

    def serve_forever(self) -> None:
        self.client.process()

    def add_event_handler(self, name: str, cb: Callable) -> None:
        self.client.add_event_handler(name, cb)

    def del_event_handler(self, name: str, cb: Callable) -> None:
        self.client.del_event_handler(name, cb)


XMPP_TO_ERR_STATUS = {
    "available": ONLINE,
    "away": AWAY,
    "dnd": DND,
    "unavailable": OFFLINE,
}


def split_identifier(txtrep: str) -> Tuple[str, str, str]:
    split_jid = txtrep.split("@", 1)
    node, domain = "@".join(split_jid[:-1]), split_jid[-1]
    if domain.find("/") != -1:
        domain, resource = domain.split("/", 1)
    else:
        resource = None

    return node, domain, resource


class XMPPBackend(ErrBot):
    room_factory = XMPPRoom
    roomoccupant_factory = XMPPRoomOccupant

    def __init__(self, config):
        super().__init__(config)
        identity = config.BOT_IDENTITY

        self.jid = identity["username"]  # backward compatibility
        self.password = identity["password"]
        self.server = identity.get("server", None)
        self.feature = config.__dict__.get("XMPP_FEATURE_MECHANISMS", {})
        self.keepalive = config.__dict__.get("XMPP_KEEPALIVE_INTERVAL", None)
        self.ca_cert = config.__dict__.get(
            "XMPP_CA_CERT_FILE", "/etc/ssl/certs/ca-certificates.crt"
        )
        self.xhtmlim = config.__dict__.get("XMPP_XHTML_IM", False)
        self.use_ipv6 = config.__dict__.get("XMPP_USE_IPV6", None)
        self.ssl_version = config.__dict__.get("XMPP_SSL_VERSION", None)

        # generic backend compatibility
        self.bot_identifier = self._build_person(self.jid)

        self.conn = self.create_connection()
        self.conn.add_event_handler("message", self.incoming_message)
        self.conn.add_event_handler("session_start", self.connected)
        self.conn.add_event_handler("disconnected", self.disconnected)
        # presence related handlers
        self.conn.add_event_handler("got_online", self.contact_online)
        self.conn.add_event_handler("got_offline", self.contact_offline)
        self.conn.add_event_handler("changed_status", self.user_changed_status)
        # MUC subject events
        self.conn.add_event_handler("groupchat_subject", self.chat_topic)
        self._room_topics = {}
        self.md_xhtml = xhtml()
        self.md_text = text()

    def create_connection(self) -> XMPPConnection:
        return XMPPConnection(
            jid=self.jid,  # textual and original representation
            password=self.password,
            feature=self.feature,
            keepalive=self.keepalive,
            ca_cert=self.ca_cert,
            server=self.server,
            use_ipv6=self.use_ipv6,
            bot=self
        )

    def _build_room_occupant(self, txtrep: str) -> XMPPRoomOccupant:
        node, domain, resource = split_identifier(txtrep)
        return self.roomoccupant_factory(
            node, domain, resource, self.query_room(node + "@" + domain)
        )

    def _build_person(self, txtrep: str) -> XMPPPerson:
        return XMPPPerson(*split_identifier(txtrep))

    def incoming_message(self, xmppmsg: dict) -> None:
        """Callback for message events"""
        if xmppmsg["type"] == "error":
            log.warning("Received error message: %s", xmppmsg)
            return

        msg = Message(xmppmsg["body"])
        if "html" in xmppmsg.keys():
            msg.html = xmppmsg["html"]
        log.debug("incoming_message from: %s", msg.frm)
        if xmppmsg["type"] == "groupchat":
            msg.frm = self._build_room_occupant(xmppmsg["from"].full)
            msg.to = msg.frm.room
        else:
            msg.frm = self._build_person(xmppmsg["from"].full)
            msg.to = self._build_person(xmppmsg["to"].full)

        msg.nick = xmppmsg["mucnick"]
        now = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
        delay = xmppmsg["delay"]._get_attr(
            "stamp"
        )  # this is a bug in sleekxmpp it should be ['from']
        msg.delayed = bool(delay and delay != now)
        self.callback_message(msg)

    def _idd_from_event(self, event) -> Union[XMPPRoomOccupant, XMPPPerson]:
        txtrep = event["from"].full
        return (
            self._build_room_occupant(txtrep)
            if "muc" in event
            else self._build_person(txtrep)
        )

    def contact_online(self, event) -> None:
        log.debug("contact_online %s.", event)
        self.callback_presence(
            Presence(identifier=self._idd_from_event(event), status=ONLINE)
        )

    def contact_offline(self, event) -> None:
        log.debug("contact_offline %s.", event)
        self.callback_presence(
            Presence(identifier=self._idd_from_event(event), status=OFFLINE)
        )

    def user_joined_chat(self, event) -> None:
        log.debug("user_join_chat %s", event)
        self.callback_presence(
            Presence(identifier=self._idd_from_event(event), status=ONLINE)
        )

    def user_left_chat(self, event) -> None:
        log.debug("user_left_chat %s", event)
        self.callback_presence(
            Presence(identifier=self._idd_from_event(event), status=OFFLINE)
        )

    def chat_topic(self, event) -> None:
        log.debug("chat_topic %s.", event)
        room = event.values["mucroom"]
        topic = event.values["subject"]
        if topic == "":
            topic = None
        self._room_topics[room] = topic
        room = XMPPRoom(event.values["mucroom"], self)
        self.callback_room_topic(room)

    def user_changed_status(self, event) -> None:
        log.debug("user_changed_status %s.", event)
        errstatus = XMPP_TO_ERR_STATUS.get(event["type"], None)
        message = event["status"]
        if not errstatus:
            errstatus = event["type"]
        self.callback_presence(
            Presence(
                identifier=self._idd_from_event(event),
                status=errstatus,
                message=message,
            )
        )

    def connected(self, data) -> None:
        """Callback for connection events"""
        self.connect_callback()

    def disconnected(self, data) -> None:
        """Callback for disconnection events"""
        self.disconnect_callback()

    def send_message(self, msg: Message) -> None:
        super().send_message(msg)

        log.debug("send_message to %s", msg.to)

        # We need to unescape the unicode characters (not the markup incompatible ones)
        mhtml = (
            xhtmlim.unescape(self.md_xhtml.convert(msg.body)) if self.xhtmlim else None
        )

        self.conn.client.send_message(
            mto=str(msg.to),
            mbody=self.md_text.convert(msg.body),
            mhtml=mhtml,
            mtype="chat" if msg.is_direct else "groupchat",
        )

    def change_presence(self, status: str = ONLINE, message: str = "") -> None:
        log.debug("Change bot status to %s, message %s.", status, message)
        self.conn.client.send_presence(pshow=status, pstatus=message)

    def serve_forever(self) -> None:
        self.conn.connect()

        try:
            self.conn.serve_forever()
        finally:
            log.debug("Trigger disconnect callback")
            self.disconnect_callback()
            log.debug("Trigger shutdown")
            self.shutdown()

    @lru_cache(IDENTIFIERS_LRU)
    def build_identifier(
        self, txtrep: str
    ) -> Union[XMPPRoomOccupant, XMPPRoom, XMPPPerson]:
        log.debug("build identifier for %s", txtrep)
        try:
            xep0030 = self.conn.client.plugin["xep_0030"]
            info = xep0030.get_info(jid=txtrep)
            disco_info = info["disco_info"]
            if disco_info:
                for category, typ, _, name in disco_info["identities"]:
                    if category == "conference":
                        log.debug("This is a room ! %s", txtrep)
                        return self.query_room(txtrep)
                    if (
                        category == "client"
                        and "http://jabber.org/protocol/muc"
                        in info["disco_info"]["features"]
                    ):
                        log.debug("This is room occupant ! %s", txtrep)
                        return self._build_room_occupant(txtrep)
        except IqError as iq:
            log.debug("xep_0030 is probably not implemented on this server. %s.", iq)
        log.debug("This is a person ! %s", txtrep)
        return self._build_person(txtrep)

    def build_reply(
        self,
        msg: Message,
        text: str = None,
        private: bool = False,
        threaded: bool = False,
    ) -> Message:
        response = self.build_message(text)
        response.frm = self.bot_identifier

        if msg.is_group and not private:
            # stripped returns the full bot@conference.domain.tld/chat_username
            # but in case of a groupchat, we should only try to send to the MUC address
            # itself (bot@conference.domain.tld)
            response.to = XMPPRoom(msg.frm.node + "@" + msg.frm.domain, self)
        elif msg.is_direct:
            # preserve from in case of a simple chat message.
            # it is either a user to user or user_in_chatroom to user case.
            # so we need resource.
            response.to = msg.frm
        elif (
            hasattr(msg.to, "person")
            and msg.to.person == self.bot_config.BOT_IDENTITY["username"]
        ):
            # This is a direct private message, not initiated through a MUC. Use
            # stripped to remove the resource so that the response goes to the
            # client with the highest priority
            response.to = XMPPPerson(msg.frm.node, msg.frm.domain, None)
        else:
            # This is a private message that was initiated through a MUC. Don't use
            # stripped here to retain the resource, else the XMPP server doesn't
            # know which user we're actually responding to.
            response.to = msg.frm
        return response

    @property
    def mode(self):
        return "xmpp"

    def rooms(self) -> List[XMPPRoom]:
        """
        Return a list of rooms the bot is currently in.

        :returns:
            A list of :class:`~errbot.backends.base.XMPPMUCRoom` instances.
        """
        xep0045 = self.conn.client.plugin["xep_0045"]
        return [XMPPRoom(room, self) for room in xep0045.get_joined_rooms()]

    def query_room(self, room) -> XMPPRoom:
        """
        Query a room for information.

        :param room:
            The JID/identifier of the room to query for.
        :returns:
            An instance of :class:`~XMPPMUCRoom`.
        """
        return XMPPRoom(room, self)

    def prefix_groupchat_reply(self, message: Message, identifier: Identifier):
        super().prefix_groupchat_reply(message, identifier)
        message.body = f"@{identifier.nick} {message.body}"

    def __hash__(self):
        return 0



import logging
import os
import sys
from json import loads
from random import randrange
from threading import Thread
from urllib.request import unquote

from OpenSSL import crypto
from webtest import TestApp
from werkzeug.serving import ThreadedWSGIServer

from errbot import BotPlugin, botcmd, webhook
from errbot.core_plugins import flask_app


def make_ssl_certificate(key_path, cert_path):
    """
    Generate a self-signed certificate

    The generated key will be written out to key_path, with the corresponding
    certificate itself being written to cert_path.
    :param cert_path: path where to write the certificate.
    :param key_path: path where to write the key.
    """
    cert = crypto.X509()
    cert.set_serial_number(randrange(1, sys.maxsize))
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(60 * 60 * 24 * 365)

    subject = cert.get_subject()
    subject.CN = "*"
    setattr(
        subject, "O", "Self-Signed Certificate for Errbot"
    )  # Pep8 annoyance workaround

    issuer = cert.get_issuer()
    issuer.CN = "Self-proclaimed Authority"
    setattr(issuer, "O", "Self-Signed")  # Pep8 annoyance workaround

    pkey = crypto.PKey()
    pkey.generate_key(crypto.TYPE_RSA, 4096)
    cert.set_pubkey(pkey)
    cert.sign(pkey, "sha256")

    f = open(cert_path, "w")
    f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"))
    f.close()

    f = open(key_path, "w")
    f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey).decode("utf-8"))
    f.close()


class Webserver(BotPlugin):
    def __init__(self, *args, **kwargs):
        self.server = None
        self.server_thread = None
        self.ssl_context = None
        self.test_app = TestApp(flask_app)
        super().__init__(*args, **kwargs)

    def get_configuration_template(self):
        return {
            "HOST": "0.0.0.0",
            "PORT": 3141,
            "SSL": {
                "enabled": False,
                "host": "0.0.0.0",
                "port": 3142,
                "certificate": "",
                "key": "",
            },
        }

    def check_configuration(self, configuration):
        # it is a pain, just assume a default config if SSL is absent or set to None
        if configuration.get("SSL", None) is None:
            configuration["SSL"] = {
                "enabled": False,
                "host": "0.0.0.0",
                "port": 3142,
                "certificate": "",
                "key": "",
            }
        super().check_configuration(configuration)

    def activate(self):
        if not self.config:
            self.log.info("Webserver is not configured. Forbid activation")
            return

        if self.server_thread and self.server_thread.is_alive():
            raise Exception(
                "Invalid state, you should not have a webserver already running."
            )
        self.server_thread = Thread(target=self.run_server, name="Webserver Thread")
        self.server_thread.start()
        self.log.debug("Webserver started.")

        super().activate()

    def deactivate(self):
        if self.server is not None:
            self.log.info("Shutting down the internal webserver.")
            self.server.shutdown()
            self.log.info("Waiting for the webserver thread to quit.")
            self.server_thread.join()
            self.log.info("Webserver shut down correctly.")
        super().deactivate()

    def run_server(self):
        try:
            host = self.config["HOST"]
            port = self.config["PORT"]
            ssl = self.config["SSL"]
            self.log.info("Starting the webserver on %s:%i", host, port)
            ssl_context = (ssl["certificate"], ssl["key"]) if ssl["enabled"] else None
            self.server = ThreadedWSGIServer(
                host,
                ssl["port"] if ssl_context else port,
                flask_app,
                ssl_context=ssl_context,
            )
            wsgi_log = logging.getLogger("werkzeug")
            wsgi_log.setLevel(self.bot_config.BOT_LOG_LEVEL)
            self.server.serve_forever()
            self.log.debug("Webserver stopped")
        except KeyboardInterrupt:
            self.log.info("Keyboard interrupt, request a global shutdown.")
            self.server.shutdown()
        except Exception:
            self.log.exception("The webserver exploded.")

    @botcmd(template="webstatus")
    def webstatus(self, msg, args):
        """
        Gives a quick status of what is mapped in the internal webserver
        """
        return {
            "rules": (((rule.rule, rule.endpoint) for rule in flask_app.url_map._rules))
        }

    @webhook
    def echo(self, incoming_request):
        """
        A simple test webhook
        """
        self.log.debug("Your incoming request is: %s", incoming_request)
        return str(incoming_request)

    @botcmd(split_args_with=" ", template="webserver")
    def webhook_test(self, _, args):
        """
            Test your webhooks from within err.

        The syntax is :
        !webhook test [relative_url] [post content]

        It triggers the notification and generate also a little test report.
        """
        url = args[0]
        content = " ".join(args[1:])

        # try to guess the content-type of what has been passed
        try:
            # try if it is plain json
            loads(content)
            contenttype = "application/json"
        except ValueError:
            # try if it is a form
            splitted = content.split("=")
            # noinspection PyBroadException
            try:
                payload = "=".join(splitted[1:])
                loads(unquote(payload))
                contenttype = "application/x-www-form-urlencoded"
            except Exception as _:
                contenttype = "text/plain"  # dunno what it is

        self.log.debug("Detected your post as : %s.", contenttype)

        response = self.test_app.post(url, params=content, content_type=contenttype)
        return {
            "url": url,
            "content": content,
            "contenttype": contenttype,
            "response": response,
        }

    @botcmd(admin_only=True)
    def generate_certificate(self, _, args):
        """
        Generate a self-signed SSL certificate for the Webserver
        """
        yield (
            "Generating a new private key and certificate. This could take a "
            "while if your system is slow or low on entropy"
        )
        key_path = os.sep.join((self.bot_config.BOT_DATA_DIR, "webserver_key.pem"))
        cert_path = os.sep.join(
            (self.bot_config.BOT_DATA_DIR, "webserver_certificate.pem")
        )
        make_ssl_certificate(key_path=key_path, cert_path=cert_path)
        yield f"Certificate successfully generated and saved in {self.bot_config.BOT_DATA_DIR}."

        suggested_config = self.config
        suggested_config["SSL"]["enabled"] = True
        suggested_config["SSL"]["host"] = suggested_config["HOST"]
        suggested_config["SSL"]["port"] = suggested_config["PORT"] + 1
        suggested_config["SSL"]["key"] = key_path
        suggested_config["SSL"]["certificate"] = cert_path
        yield "To enable SSL with this certificate, the following config is recommended:"
        yield f"{suggested_config!r}"



import errno
import os
import re
import sys
from configparser import ConfigParser
from typing import List, Optional

import jinja2

from errbot.version import VERSION


def new_plugin_wizard(directory: Optional[str] = None) -> None:
    """
    Start the wizard to create a new plugin in the current working directory.
    """
    if directory is None:
        print("This wizard will create a new plugin for you in the current directory.")
        directory = os.getcwd()
    else:
        print(f"This wizard will create a new plugin for you in '{directory}'.")

    if os.path.exists(directory) and not os.path.isdir(directory):
        print(f"Error: The path '{directory}' exists but it isn't a directory")
        sys.exit(1)

    name = ask(
        "What should the name of your new plugin be?",
        validation_regex=r"^[a-zA-Z][a-zA-Z0-9 _-]*$",
    ).strip()
    module_name = name.lower().replace(" ", "_")
    directory_name = name.lower().replace(" ", "-")
    class_name = "".join([s.capitalize() for s in name.lower().split(" ")])

    description = ask(
        "What may I use as a short (one-line) description of your plugin?"
    )
    python_version = "3"
    errbot_min_version = ask(
        f"Which minimum version of errbot will your plugin work with?\n"
        f"Leave blank to support any version or input CURRENT to select "
        f"the current version {VERSION}."
    ).strip()
    if errbot_min_version.upper() == "CURRENT":
        errbot_min_version = VERSION
    errbot_max_version = ask(
        f"Which maximum version of errbot will your plugin work with?\n"
        f"Leave blank to support any version or input CURRENT to select "
        f"the current version {VERSION}."
    ).strip()
    if errbot_max_version.upper() == "CURRENT":
        errbot_max_version = VERSION

    plug = ConfigParser()
    plug["Core"] = {
        "Name": name,
        "Module": module_name,
    }
    plug["Documentation"] = {
        "Description": description,
    }
    plug["Python"] = {
        "Version": python_version,
    }

    if errbot_max_version != "" or errbot_min_version != "":
        plug["Errbot"] = {}
        if errbot_min_version != "":
            plug["Errbot"]["Min"] = errbot_min_version
        if errbot_max_version != "":
            plug["Errbot"]["Max"] = errbot_max_version

    plugin_path = directory
    plugfile_path = os.path.join(plugin_path, module_name + ".plug")
    pyfile_path = os.path.join(plugin_path, module_name + ".py")

    try:
        os.makedirs(plugin_path, mode=0o700)
    except IOError as e:
        if e.errno != errno.EEXIST:
            raise

    if os.path.exists(plugfile_path) or os.path.exists(pyfile_path):
        path = os.path.join(directory, f"{module_name}.{{py,plug}}")
        ask(
            f"Warning: A plugin with this name was already found at {path}\n"
            f"If you continue, these will be overwritten.\n"
            f"Press Ctrl+C to abort now or type in 'overwrite' to confirm overwriting of these files.",
            valid_responses=["overwrite"],
        )

    with open(plugfile_path, "w") as f:
        plug.write(f)

    with open(pyfile_path, "w") as f:
        f.write(render_plugin(locals()))

    print(f"Success! You'll find your new plugin at '{plugfile_path}'")
    print(
        "(Don't forget to include a LICENSE file if you are going to publish your plugin)."
    )


def ask(
    question: str,
    valid_responses: Optional[List[str]] = None,
    validation_regex: Optional[str] = None,
) -> Optional[str]:
    """
    Ask the user for some input. If valid_responses is supplied, the user
    must respond with something present in this list.
    """
    response = None
    print(question)
    while True:
        response = input("> ")
        if valid_responses is not None:
            assert isinstance(valid_responses, list)
            if response in valid_responses:
                break
            else:
                print(f"Bad input: Please answer one of: {', '.join(valid_responses)}")
        elif validation_regex is not None:
            m = re.search(validation_regex, response)
            if m is None:
                print(
                    f"Bad input: Please respond with something matching this regex: {validation_regex}"
                )
            else:
                break
        else:
            break
    return response


def render_plugin(values) -> jinja2.Template:
    """
    Render the Jinja template for the plugin with the given values.
    """
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(
            os.path.join(os.path.dirname(__file__), "templates")
        ),
        auto_reload=False,
        keep_trailing_newline=True,
        autoescape=True,
    )
    template = env.get_template("new_plugin.py.tmpl")
    return template.render(**values)


if __name__ == "__main__":
    try:
        new_plugin_wizard()
    except KeyboardInterrupt:
        sys.exit(1)



import importlib
import logging
import sys
import warnings
from os import makedirs, path
from typing import Callable, Optional

from errbot.backend_plugin_manager import BackendPluginManager
from errbot.core import ErrBot
from errbot.logs import format_logs
from errbot.plugin_manager import BotPluginManager
from errbot.repo_manager import BotRepoManager
from errbot.storage.base import StoragePluginBase
from errbot.utils import PLUGINS_SUBDIR

log = logging.getLogger(__name__)

HERE = path.dirname(path.abspath(__file__))
CORE_BACKENDS = path.join(HERE, "backends")
CORE_STORAGE = path.join(HERE, "storage")

PLUGIN_DEFAULT_INDEX = "https://errbot.io/repos.json"


def bot_config_defaults(config: object) -> None:
    if not hasattr(config, "ACCESS_CONTROLS_DEFAULT"):
        config.ACCESS_CONTROLS_DEFAULT = {}
    if not hasattr(config, "ACCESS_CONTROLS"):
        config.ACCESS_CONTROLS = {}
    if not hasattr(config, "HIDE_RESTRICTED_COMMANDS"):
        config.HIDE_RESTRICTED_COMMANDS = False
    if not hasattr(config, "HIDE_RESTRICTED_ACCESS"):
        config.HIDE_RESTRICTED_ACCESS = False
    if not hasattr(config, "BOT_PREFIX_OPTIONAL_ON_CHAT"):
        config.BOT_PREFIX_OPTIONAL_ON_CHAT = False
    if not hasattr(config, "BOT_PREFIX"):
        config.BOT_PREFIX = "!"
    if not hasattr(config, "BOT_ALT_PREFIXES"):
        config.BOT_ALT_PREFIXES = ()
    if not hasattr(config, "BOT_ALT_PREFIX_SEPARATORS"):
        config.BOT_ALT_PREFIX_SEPARATORS = ()
    if not hasattr(config, "BOT_ALT_PREFIX_CASEINSENSITIVE"):
        config.BOT_ALT_PREFIX_CASEINSENSITIVE = False
    if not hasattr(config, "DIVERT_TO_PRIVATE"):
        config.DIVERT_TO_PRIVATE = ()
    if not hasattr(config, "DIVERT_TO_THREAD"):
        config.DIVERT_TO_THREAD = ()
    if not hasattr(config, "MESSAGE_SIZE_LIMIT"):
        config.MESSAGE_SIZE_LIMIT = None  # No user limit declared.
    if not hasattr(config, "GROUPCHAT_NICK_PREFIXED"):
        config.GROUPCHAT_NICK_PREFIXED = False
    if not hasattr(config, "AUTOINSTALL_DEPS"):
        config.AUTOINSTALL_DEPS = True
    if not hasattr(config, "SUPPRESS_CMD_NOT_FOUND"):
        config.SUPPRESS_CMD_NOT_FOUND = False
    if not hasattr(config, "BOT_ASYNC"):
        config.BOT_ASYNC = True
    if not hasattr(config, "BOT_ASYNC_POOLSIZE"):
        config.BOT_ASYNC_POOLSIZE = 10
    if not hasattr(config, "CHATROOM_PRESENCE"):
        config.CHATROOM_PRESENCE = ()
    if not hasattr(config, "CHATROOM_RELAY"):
        config.CHATROOM_RELAY = ()
    if not hasattr(config, "REVERSE_CHATROOM_RELAY"):
        config.REVERSE_CHATROOM_RELAY = ()
    if not hasattr(config, "CHATROOM_FN"):
        config.CHATROOM_FN = "Errbot"
    if not hasattr(config, "TEXT_DEMO_MODE"):
        config.TEXT_DEMO_MODE = True
    if not hasattr(config, "BOT_ADMINS"):
        raise ValueError("BOT_ADMINS missing from config.py.")
    if not hasattr(config, "TEXT_COLOR_THEME"):
        config.TEXT_COLOR_THEME = "light"
    if not hasattr(config, "BOT_ADMINS_NOTIFICATIONS"):
        config.BOT_ADMINS_NOTIFICATIONS = config.BOT_ADMINS


def setup_bot(
    backend_name: str,
    logger: logging.Logger,
    config: object,
    restore: Optional[str] = None,
) -> ErrBot:
    # from here the environment is supposed to be set (daemon / non daemon,
    # config.py in the python path )

    bot_config_defaults(config)

    if hasattr(config, "BOT_LOG_FORMATTER"):
        format_logs(formatter=config.BOT_LOG_FORMATTER)
    else:
        format_logs(theme_color=config.TEXT_COLOR_THEME)

    if hasattr(config, "BOT_LOG_FILE") and config.BOT_LOG_FILE:
        hdlr = logging.FileHandler(config.BOT_LOG_FILE)
        hdlr.setFormatter(
            logging.Formatter("%(asctime)s %(levelname)-8s %(name)-25s %(message)s")
        )
        logger.addHandler(hdlr)

    if hasattr(config, "BOT_LOG_SENTRY") and config.BOT_LOG_SENTRY:
        sentry_integrations = []

        try:
            import sentry_sdk
            from sentry_sdk.integrations.logging import LoggingIntegration

        except ImportError:
            log.exception(
                "You have BOT_LOG_SENTRY enabled, but I couldn't import modules "
                "needed for Sentry integration. Did you install sentry-sdk? "
                "(See https://docs.sentry.io/platforms/python for installation instructions)"
            )
            exit(-1)

        sentry_logging = LoggingIntegration(
            level=config.SENTRY_LOGLEVEL, event_level=config.SENTRY_EVENTLEVEL
        )

        sentry_integrations.append(sentry_logging)

        if hasattr(config, "BOT_LOG_SENTRY_FLASK") and config.BOT_LOG_SENTRY_FLASK:
            try:
                from sentry_sdk.integrations.flask import FlaskIntegration
            except ImportError:
                log.exception(
                    "You have BOT_LOG_SENTRY enabled, but I couldn't import modules "
                    "needed for Sentry integration. Did you install sentry-sdk[flask]? "
                    "(See https://docs.sentry.io/platforms/python/flask for installation instructions)"
                )
                exit(-1)

            sentry_integrations.append(FlaskIntegration())

        sentry_options = getattr(config, "SENTRY_OPTIONS", {})
        if hasattr(config, "SENTRY_TRANSPORT") and isinstance(
            config.SENTRY_TRANSPORT, tuple
        ):
            warnings.warn(
                "SENTRY_TRANSPORT is deprecated. Please use SENTRY_OPTIONS instead.",
                DeprecationWarning,
            )
            try:
                mod = importlib.import_module(config.SENTRY_TRANSPORT[1])
                transport = getattr(mod, config.SENTRY_TRANSPORT[0])
                sentry_options["transport"] = transport
            except ImportError:
                log.exception(
                    f"Unable to import selected SENTRY_TRANSPORT - {config.SENTRY_TRANSPORT}"
                )
                exit(-1)
        # merge options dict with dedicated SENTRY_DSN setting
        sentry_kwargs = {
            **sentry_options,
            **{"dsn": config.SENTRY_DSN, "integrations": sentry_integrations},
        }
        sentry_sdk.init(**sentry_kwargs)

    logger.setLevel(config.BOT_LOG_LEVEL)

    storage_plugin = get_storage_plugin(config)

    # init the botplugin manager
    botplugins_dir = path.join(config.BOT_DATA_DIR, PLUGINS_SUBDIR)
    if not path.exists(botplugins_dir):
        makedirs(botplugins_dir, mode=0o755)

    plugin_indexes = getattr(config, "BOT_PLUGIN_INDEXES", (PLUGIN_DEFAULT_INDEX,))
    if isinstance(plugin_indexes, str):
        plugin_indexes = (plugin_indexes,)

    # Extra backend is expected to be a list type, convert string to list.
    extra_backend = getattr(config, "BOT_EXTRA_BACKEND_DIR", [])
    if isinstance(extra_backend, str):
        extra_backend = [extra_backend]

    backendpm = BackendPluginManager(
        config, "errbot.backends", backend_name, ErrBot, CORE_BACKENDS, extra_backend
    )

    log.info(f"Found Backend plugin: {backendpm.plugin_info.name}")

    repo_manager = BotRepoManager(storage_plugin, botplugins_dir, plugin_indexes)

    try:
        bot = backendpm.load_plugin()
        botpm = BotPluginManager(
            storage_plugin,
            config.BOT_EXTRA_PLUGIN_DIR,
            config.AUTOINSTALL_DEPS,
            getattr(config, "CORE_PLUGINS", None),
            lambda name, clazz: clazz(bot, name),
            getattr(config, "PLUGINS_CALLBACK_ORDER", (None,)),
        )
        bot.attach_storage_plugin(storage_plugin)
        bot.attach_repo_manager(repo_manager)
        bot.attach_plugin_manager(botpm)
        bot.initialize_backend_storage()

        # restore the bot from the restore script
        if restore:
            # Prepare the context for the restore script
            if "repos" in bot:
                log.fatal("You cannot restore onto a non empty bot.")
                sys.exit(-1)
            log.info(f"**** RESTORING the bot from {restore}")
            restore_bot_from_backup(restore, bot=bot, log=log)
            print("Restore complete. You can restart the bot normally")
            sys.exit(0)

        errors = bot.plugin_manager.update_plugin_places(
            repo_manager.get_all_repos_paths()
        )
        if errors:
            startup_errors = "\n".join(errors.values())
            log.error("Some plugins failed to load:\n%s", startup_errors)
            bot._plugin_errors_during_startup = startup_errors
        return bot
    except Exception:
        log.exception("Unable to load or configure the backend.")
        exit(-1)


def restore_bot_from_backup(backup_filename: str, *, bot, log: logging.Logger):
    """Restores the given bot by executing the 'backup' script.

    The backup file is a python script which manually execute a series of commands on the bot
    to restore it to its previous state.

    :param backup_filename: the full path to the backup script.
    :param bot: the bot instance to restore
    :param log: logger to use during the restoration process
    """
    with open(backup_filename) as f:
        exec(f.read(), {"log": log, "bot": bot})
    bot.close_storage()


def get_storage_plugin(config: object) -> Callable:
    """
    Find and load the storage plugin
    :param config: the bot configuration.
    :return: the storage plugin
    """
    storage_name = getattr(config, "STORAGE", "Shelf")
    extra_storage_plugins_dir = getattr(config, "BOT_EXTRA_STORAGE_PLUGINS_DIR", None)
    spm = BackendPluginManager(
        config,
        "errbot.storage",
        storage_name,
        StoragePluginBase,
        CORE_STORAGE,
        extra_storage_plugins_dir,
    )
    log.info(f"Found Storage plugin: {spm.plugin_info.name}.")
    return spm.load_plugin()


def bootstrap(
    bot_class, logger: logging.Logger, config: object, restore: Optional[str] = None
) -> None:
    """
    Main starting point of Errbot.

    :param bot_class: The backend class inheriting from Errbot you want to start.
    :param logger: The logger you want to use.
    :param config: The config.py module.
    :param restore: Start Errbot in restore mode (from a backup).
    """
    bot = setup_bot(bot_class, logger, config, restore)
    log.debug(f"Start serving commands from the {bot.mode} backend.")
    bot.serve_forever()



import inspect
import sys
from configparser import ConfigParser
from configparser import Error as ConfigParserError
from dataclasses import dataclass
from importlib._bootstrap import module_from_spec
from importlib._bootstrap_external import spec_from_file_location
from pathlib import Path
from typing import List, Tuple, Type

from errbot.utils import version2tuple

VersionType = Tuple[int, int, int]


@dataclass
class PluginInfo:
    name: str
    module: str
    doc: str
    core: bool
    python_version: VersionType
    errbot_minversion: VersionType
    errbot_maxversion: VersionType
    dependencies: List[str]
    location: Path = None

    @staticmethod
    def load(plugfile_path: Path) -> "PluginInfo":
        with plugfile_path.open(encoding="utf-8") as plugfile:
            return PluginInfo.load_file(plugfile, plugfile_path)

    @staticmethod
    def load_file(plugfile, location: Path) -> "PluginInfo":
        cp = ConfigParser()
        cp.read_file(plugfile)
        pi = PluginInfo.parse(cp)
        pi.location = location
        return pi

    @staticmethod
    def parse(config: ConfigParser) -> "PluginInfo":
        """
        Throws ConfigParserError with a meaningful message if the ConfigParser doesn't contain the minimal
         information required.
        """
        name = config.get("Core", "Name")
        module = config.get("Core", "Module")
        core = config.get("Core", "Core", fallback="false").lower() == "true"
        doc = config.get("Documentation", "Description", fallback=None)

        python_version = config.get("Python", "Version", fallback=None)
        # Old format backward compatibility
        if python_version:
            if python_version in ("2+", "3"):
                python_version = (3, 0, 0)
            elif python_version == "2":
                python_version = (2, 0, 0)
            else:
                try:
                    python_version = tuple(
                        version2tuple(python_version)[0:3]
                    )  # We can ignore the alpha/beta part.
                except ValueError as ve:
                    raise ConfigParserError(
                        f"Invalid Python Version format: {python_version} ({ve})"
                    )

        min_version = config.get("Errbot", "Min", fallback=None)
        max_version = config.get("Errbot", "Max", fallback=None)
        try:
            if min_version:
                min_version = version2tuple(min_version)
        except ValueError as ve:
            raise ConfigParserError(
                f"Invalid Errbot min version format: {min_version} ({ve})"
            )

        try:
            if max_version:
                max_version = version2tuple(max_version)
        except ValueError as ve:
            raise ConfigParserError(
                f"Invalid Errbot max version format: {max_version} ({ve})"
            )
        depends_on = config.get("Core", "DependsOn", fallback=None)
        deps = [name.strip() for name in depends_on.split(",")] if depends_on else []

        return PluginInfo(
            name, module, doc, core, python_version, min_version, max_version, deps
        )

    def load_plugin_classes(self, base_module_name: str, baseclass: Type):
        # load the module
        module_name = base_module_name + "." + self.module
        spec = spec_from_file_location(
            module_name, self.location.parent / (self.module + ".py")
        )
        modu1e = module_from_spec(spec)
        spec.loader.exec_module(modu1e)
        sys.modules[module_name] = modu1e

        # introspect the modules to find plugin classes
        def is_plugin(member):
            return (
                inspect.isclass(member)
                and issubclass(member, baseclass)
                and member != baseclass
            )

        plugin_classes = inspect.getmembers(modu1e, is_plugin)
        return plugin_classes



import gc
import os
import signal
from datetime import datetime

from errbot import BotPlugin, arg_botcmd, botcmd
from errbot.utils import format_timedelta, global_restart


class Health(BotPlugin):
    @botcmd(template="status")
    def status(self, msg, args):
        """If I am alive I should be able to respond to this one"""
        plugins_statuses = self.status_plugins(msg, args)
        loads = self.status_load(msg, args)
        gc = self.status_gc(msg, args)

        return {
            "plugins_statuses": plugins_statuses["plugins_statuses"],
            "loads": loads["loads"],
            "gc": gc["gc"],
        }

    @botcmd(template="status_load")
    def status_load(self, _, args):
        """shows the load status"""
        try:
            from posix import getloadavg

            loads = getloadavg()
        except Exception:
            loads = None

        return {"loads": loads}

    @botcmd(template="status_gc")
    def status_gc(self, _, args):
        """shows the garbage collection details"""
        return {"gc": gc.get_count()}

    @botcmd(template="status_plugins")
    def status_plugins(self, _, args):
        """shows the plugin status"""
        pm = self._bot.plugin_manager
        all_blacklisted = pm.get_blacklisted_plugin()
        all_loaded = pm.get_all_active_plugin_names()
        all_attempted = sorted(pm.plugin_infos.keys())
        plugins_statuses = []
        for name in all_attempted:
            if name in all_blacklisted:
                if name in all_loaded:
                    plugins_statuses.append(("BA", name))
                else:
                    plugins_statuses.append(("BD", name))
            elif name in all_loaded:
                plugins_statuses.append(("A", name))
            elif (
                pm.get_plugin_obj_by_name(name) is not None
                and pm.get_plugin_obj_by_name(name).get_configuration_template()
                is not None
                and pm.get_plugin_configuration(name) is None
            ):
                plugins_statuses.append(("C", name))
            else:
                plugins_statuses.append(("D", name))

        return {"plugins_statuses": plugins_statuses}

    @botcmd
    def uptime(self, _, args):
        """Return the uptime of the bot"""
        u = format_timedelta(datetime.now() - self._bot.startup_time)
        since = self._bot.startup_time.strftime("%A, %b %d at %H:%M")
        return f"I've been up for {u} (since {since})."

    # noinspection PyUnusedLocal
    @botcmd(admin_only=True)
    def restart(self, msg, args):
        """Restart the bot."""
        self.send(msg.frm, "Deactivating all the plugins...")
        self._bot.plugin_manager.deactivate_all_plugins()
        self.send(msg.frm, "Restarting")
        self._bot.shutdown()
        global_restart()
        return "I'm restarting..."

    # noinspection PyUnusedLocal
    @arg_botcmd(
        "--confirm",
        dest="confirmed",
        action="store_true",
        help="confirm you want to shut down",
        admin_only=True,
    )
    @arg_botcmd(
        "--kill",
        dest="kill",
        action="store_true",
        help="kill the bot instantly, don't shut down gracefully",
        admin_only=True,
    )
    def shutdown(self, msg, confirmed, kill):
        """
        Shutdown the bot.

        Useful when the things are going crazy and you don't have access to the machine.
        """
        if not confirmed:
            yield "Please provide `--confirm` to confirm you really want me to shut down."
            return

        if kill:
            yield "Killing myself right now!"
            os.kill(os.getpid(), signal.SIGKILL)
        else:
            yield "Roger that. I am shutting down."
            os.kill(os.getpid(), signal.SIGINT)



import copyreg
import logging
import re
import sys
from time import sleep
from typing import BinaryIO, List, Optional, Union

from ansi.color import fg, fx
from markdown import Markdown
from markdown.extensions.extra import ExtraExtension
from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import get_lexer_by_name

from errbot.backends.base import (
    OFFLINE,
    ONLINE,
    Identifier,
    Message,
    Person,
    Presence,
    Room,
    RoomOccupant,
    Stream,
)
from errbot.core import ErrBot
from errbot.logs import console_hdlr
from errbot.rendering import ansi, imtext, text, xhtml
from errbot.rendering.ansiext import ANSI_CHRS, AnsiExtension, enable_format

log = logging.getLogger(__name__)

ENCODING_INPUT = sys.stdin.encoding
ANSI = hasattr(sys.stderr, "isatty") and sys.stderr.isatty()

enable_format("borderless", ANSI_CHRS, borders=False)


def borderless_ansi() -> Markdown:
    """This makes a converter from markdown to ansi (console) format.
    It can be called like this:
    from errbot.rendering import ansi
    md_converter = ansi()  # you need to cache the converter

    ansi_txt = md_converter.convert(md_txt)
    """
    md = Markdown(
        output_format="borderless", extensions=[ExtraExtension(), AnsiExtension()]
    )
    md.stripTopLevelTags = False
    return md


class TextPerson(Person):
    """
    Simple Person implementation which represents users as simple text strings.
    """

    def __init__(self, person, client=None, nick=None, fullname=None):
        self._person = person
        self._client = client
        self._nick = nick
        self._fullname = fullname
        self._email = ""

    @property
    def person(self) -> Person:
        return self._person

    @property
    def client(self) -> str:
        return self._client

    @property
    def nick(self) -> str:
        return self._nick

    @property
    def fullname(self) -> str:
        return self._fullname

    @property
    def email(self) -> str:
        return self._email

    @property
    def aclattr(self) -> str:
        return str(self)

    def __str__(self):
        return "@" + self._person

    def __eq__(self, other):
        if not isinstance(other, Person):
            return False
        return self.person == other.person

    def __hash__(self):
        return self.person.__hash__()


class TextRoom(Room):
    def __init__(self, name: str, bot: ErrBot):
        self._topic = ""
        self._joined = False
        self.name = name
        self._bot = bot

        # get the bot username
        bot_config = self._bot.bot_config
        default_bot_name = "@errbot"
        if hasattr(bot_config, "BOT_IDENTITY"):
            bot_name = bot_config.BOT_IDENTITY.get("username", default_bot_name)
        else:
            bot_name = default_bot_name

        # fill up the room with a coherent set of identities.
        self._occupants = [
            TextOccupant("somebody", self),
            TextOccupant(TextPerson(bot.bot_config.BOT_ADMINS[0]), self),
            TextOccupant(bot_name, self),
        ]

    def join(
        self, username: Optional[str] = None, password: Optional[str] = None
    ) -> None:
        self._joined = True

    def leave(self, reason: Optional[str] = None) -> None:
        self._joined = False

    def create(self) -> None:
        self._joined = True

    def destroy(self) -> None:
        self._joined = False

    @property
    def exists(self) -> bool:
        return True

    @property
    def joined(self) -> bool:
        return self._joined

    @property
    def topic(self) -> str:
        return self._topic

    @topic.setter
    def topic(self, topic: str) -> None:
        self._topic = topic

    @property
    def occupants(self) -> List["TextOccupant"]:
        return self._occupants

    def invite(self, *args):
        pass

    def __str__(self):
        return "#" + self.name

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return self.name.__hash__()


class TextOccupant(TextPerson, RoomOccupant):
    def __init__(self, person, room):
        super().__init__(person)
        self._room = room

    @property
    def room(self) -> TextRoom:
        return self._room

    def __str__(self):
        return f"#{self._room.name}/{self._person.person}"

    def __eq__(self, other):
        return self.person == other.person and self.room == other.room

    def __hash__(self):
        return self.person.__hash__() + self.room.__hash__()


INTRO = """
---
You start as a **bot admin in a one-on-one conversation** with the bot.

### Context of the chat

- Use `!inroom`{:color='blue'} to switch to a room conversation.
- Use `!inperson`{:color='blue'} to switch back to a one-on-one conversation.
- Use `!asuser`{:color='green'} to talk as a normal user.
- Use `!asadmin`{:color='red'} to switch back as a bot admin.

### Preferences

- Use `!ml`{:color='yellow'} to flip on/off the multiline mode (Enter twice at the end to send).

---
"""


class TextBackend(ErrBot):
    def __init__(self, config):
        super().__init__(config)
        log.debug("Text Backend Init.")

        if (
            hasattr(self.bot_config, "BOT_IDENTITY")
            and "username" in self.bot_config.BOT_IDENTITY
        ):
            self.bot_identifier = self.build_identifier(
                self.bot_config.BOT_IDENTITY["username"]
            )
        else:
            # Just a default identity for the bot if nothing has been specified.
            self.bot_identifier = self.build_identifier("@errbot")

        log.debug("Bot username set at %s.", self.bot_identifier)
        self._inroom = False
        self._rooms = []
        self._multiline = False

        self.demo_mode = (
            self.bot_config.TEXT_DEMO_MODE
            if hasattr(self.bot_config, "TEXT_DEMO_MODE")
            else False
        )
        if not self.demo_mode:
            self.md_html = xhtml()  # for more debug feedback on md
            self.md_text = text()  # for more debug feedback on md
            self.md_borderless_ansi = borderless_ansi()
            self.md_im = imtext()
            self.md_lexer = get_lexer_by_name("md", stripall=True)

        self.md_ansi = ansi()
        self.html_lexer = get_lexer_by_name("html", stripall=True)
        self.terminal_formatter = Terminal256Formatter(style="paraiso-dark")
        self.user = self.build_identifier(self.bot_config.BOT_ADMINS[0])
        self._register_identifiers_pickling()

    @staticmethod
    def _unpickle_identifier(identifier_str):
        return TextBackend.__build_identifier(identifier_str)

    @staticmethod
    def _pickle_identifier(identifier):
        return TextBackend._unpickle_identifier, (str(identifier),)

    def _register_identifiers_pickling(self):
        """
        Register identifiers pickling.
        """
        TextBackend.__build_identifier = self.build_identifier
        for cls in (TextPerson, TextOccupant, TextRoom):
            copyreg.pickle(
                cls, TextBackend._pickle_identifier, TextBackend._unpickle_identifier
            )

    def serve_forever(self) -> None:
        self.readline_support()

        if not self._rooms:
            # artificially join a room if None were specified.
            self.query_room("#testroom").join()

        if self.demo_mode:
            # disable the console logging once it is serving in demo mode.
            root = logging.getLogger()
            root.removeHandler(console_hdlr)
            root.addHandler(logging.NullHandler())
        self.connect_callback()  # notify that the connection occured
        self.callback_presence(Presence(identifier=self.user, status=ONLINE))

        self.send_message(Message(INTRO))

        try:
            while True:
                if self._inroom:
                    frm = TextOccupant(self.user, self.rooms()[0])
                    to = self.rooms()[0]
                else:
                    frm = self.user
                    to = self.bot_identifier

                print()
                full_msg = ""
                while True:
                    prompt = "[␍] " if full_msg else ">>> "
                    if ANSI or self.demo_mode:
                        color = (
                            fg.red
                            if self.user.person in self.bot_config.BOT_ADMINS[0]
                            else fg.green
                        )
                        prompt = f"{color}[{frm} ➡ {to}] {fg.cyan}{prompt}{fx.reset}"
                        entry = input(prompt)
                    else:
                        entry = input(f"[{frm} ➡ {to}] {prompt}")

                    if not self._multiline:
                        full_msg = entry
                        break

                    if not entry:
                        break

                    full_msg += entry + "\n"

                msg = Message(full_msg)
                msg.frm = frm
                msg.to = to

                self.callback_message(msg)

                mentioned = [
                    self.build_identifier(word)
                    for word in re.findall(r"(?<=\s)@[\w]+", entry)
                ]
                if mentioned:
                    self.callback_mention(msg, mentioned)

                sleep(0.5)
        except EOFError:
            pass
        except KeyboardInterrupt:
            pass
        finally:
            # simulate some real presence
            self.callback_presence(Presence(identifier=self.user, status=OFFLINE))
            log.debug("Trigger disconnect callback")
            self.disconnect_callback()
            log.debug("Trigger shutdown")
            self.shutdown()

    def readline_support(self) -> None:
        try:
            # Load readline for better editing/history behaviour
            import readline

            # Implement a simple completer for commands
            def completer(txt, state):
                options = (
                    [i for i in self.all_commands if i.startswith(txt)]
                    if txt
                    else list(self.all_commands.keys())
                )
                if state < len(options):
                    return options[state]

            readline.parse_and_bind("tab: complete")
            readline.set_completer(completer)

        except ImportError:
            # Readline is Unix-only
            log.debug("Python readline module is not available")

    def send_message(self, msg: Message) -> None:
        if self.demo_mode:
            print(self.md_ansi.convert(msg.body))
        else:
            bar = "\n╌╌[{mode}]" + ("╌" * 60)
            super().send_message(msg)
            print(bar.format(mode="MD  "))
            if ANSI:
                print(highlight(msg.body, self.md_lexer, self.terminal_formatter))
            else:
                print(msg.body)
            print(bar.format(mode="HTML"))
            html = self.md_html.convert(msg.body)
            if ANSI:
                print(highlight(html, self.html_lexer, self.terminal_formatter))
            else:
                print(html)
            print(bar.format(mode="TEXT"))
            print(self.md_text.convert(msg.body))
            print(bar.format(mode="IM  "))
            print(self.md_im.convert(msg.body))
            if ANSI:
                print(bar.format(mode="ANSI"))
                print(self.md_ansi.convert(msg.body))
                print(bar.format(mode="BORDERLESS"))
                print(self.md_borderless_ansi.convert(msg.body))
            print("\n\n")

    def add_reaction(self, msg: Message, reaction: str) -> None:
        # this is like the Slack backend's add_reaction
        self._react("+", msg, reaction)

    def remove_reaction(self, msg: Message, reaction: str) -> None:
        self._react("-", msg, reaction)

    def _react(self, sign: str, msg: Message, reaction: str) -> None:
        self.send(msg.frm, f"reaction {sign}:{reaction}:", in_reply_to=msg)

    def change_presence(self, status: str = ONLINE, message: str = "") -> None:
        log.debug("*** Changed presence to [%s] %s", status, message)

    def build_identifier(
        self, text_representation: str
    ) -> Union[TextOccupant, TextRoom, TextPerson]:
        if text_representation.startswith("#"):
            rem = text_representation[1:]
            if "/" in text_representation:
                room, person = rem.split("/")
                return TextOccupant(TextPerson(person), TextRoom(room, self))
            return self.query_room("#" + rem)
        if not text_representation.startswith("@"):
            raise ValueError(
                "An identifier for the Text backend needs to start with # for a room or @ for a person."
            )
        return TextPerson(text_representation[1:])

    def build_reply(
        self,
        msg: Message,
        text: str = None,
        private: bool = False,
        threaded: bool = False,
    ) -> Message:
        response = self.build_message(text)
        response.frm = self.bot_identifier
        if private:
            response.to = msg.frm
        else:
            response.to = msg.frm.room if isinstance(msg.frm, RoomOccupant) else msg.frm
        return response

    @property
    def mode(self):
        return "text"

    def query_room(self, room: str) -> TextRoom:
        if not room.startswith("#"):
            raise ValueError("A Room name must start by #.")
        text_room = TextRoom(room[1:], self)
        if text_room not in self._rooms:
            self._rooms.insert(0, text_room)
        else:
            self._rooms.insert(0, self._rooms.pop(self._rooms.index(text_room)))
        return text_room

    def rooms(self) -> List[TextRoom]:
        return self._rooms

    def prefix_groupchat_reply(self, message: Message, identifier: Identifier):
        message.body = f"{identifier.person} {message.body}"

    def send_stream_request(
        self,
        user: Identifier,
        fsource: BinaryIO,
        name: str = None,
        size: int = None,
        stream_type: str = None,
    ) -> Stream:
        """
        Starts a file transfer. For Slack, the size and stream_type are unsupported

        :param user: is the identifier of the person you want to send it to.
        :param fsource: is a file object you want to send.
        :param name: is an optional filename for it.
        :param size: not supported in Slack backend
        :param stream_type: not supported in Slack backend

        :return Stream: object on which you can monitor the progress of it.
        """
        stream = Stream(user, fsource, name, size, stream_type)
        log.debug(
            "Requesting upload of %s to %s (size hint: %d, stream type: %s).",
            stream.name,
            stream.identifier,
            size,
            stream_type,
        )
        return stream
